#include <digitalWriteFast.h>

//#define W_RAM
#define W_FLASH

//PORTC -> A0-A7
//PORTA -> A8-A15
//PORTB -> A16-A18
//PORTL -> D0-D7
//41 -> /CE
//40 -> /OE
//39 -> /WE

#define CTL_WH 7
#define CTL_WL 6
#define CTL_IPE 10
#define CTL_IPH 5
#define CTL_IPL 4
#define CTL_SEG 11

#define CTL_WH_EN 3
#define CTL_WL_EN 2

#define CTL_RH 9
#define CTL_RL 8

#define BTN_WRITE 14
#define BTN_READ 18
#define LED_EXT 15
#define HALT_EXT 16

#define FLASHSEL_B1 A8
#define FLASHSEL_B0 A9

uint32_t flashsel_offset = 0UL;


#define PULSE_LEN 2
//#define delayMicroseconds delay

void pulseHigh(int pin) {
  digitalWrite(pin, HIGH);
  delayMicroseconds(PULSE_LEN / 2);
  digitalWrite(pin, LOW);
  delayMicroseconds(PULSE_LEN / 2);
}

void pulseLow(int pin) {
  digitalWrite(pin, LOW);
  delayMicroseconds(PULSE_LEN / 2);
  digitalWrite(pin, HIGH);
  delayMicroseconds(PULSE_LEN / 2);
}

void pulseLow2(int pin, int pin2) {
  digitalWrite(pin, LOW);
  digitalWrite(pin2, LOW);
  delayMicroseconds(PULSE_LEN / 2);
  digitalWrite(pin, HIGH);
  digitalWrite(pin2, HIGH);
  delayMicroseconds(PULSE_LEN / 2);
}

void out(int pin, int data) {
  DDRF = 255;

  PORTF = data;

  if (pin == CTL_IPH || pin == CTL_IPL) {
    delayMicroseconds(PULSE_LEN / 2);
    pulseHigh(pin);
  } else if (pin == CTL_WH || pin == CTL_WL || pin == CTL_SEG || pin == CTL_IPE) {
    if (pin == CTL_WH) {
      delayMicroseconds(PULSE_LEN / 2);
      digitalWrite(CTL_WH_EN, LOW);
      digitalWrite(CTL_RH, LOW);
      delayMicroseconds(PULSE_LEN / 2);
      pulseLow(pin);
      digitalWrite(CTL_RH, HIGH);
      digitalWrite(CTL_WH_EN, HIGH);
      delayMicroseconds(PULSE_LEN / 2);
    } else if (pin == CTL_WL) {
      delayMicroseconds(PULSE_LEN / 2);
      digitalWrite(CTL_WL_EN, LOW);
      digitalWrite(CTL_RL, LOW);
      delayMicroseconds(PULSE_LEN / 2);
      pulseLow(pin);
      digitalWrite(CTL_RL, HIGH);
      digitalWrite(CTL_WL_EN, HIGH);
      delayMicroseconds(PULSE_LEN / 2);
    } else {
      delayMicroseconds(PULSE_LEN / 2);
      pulseLow(pin);
    }
  }

  DDRF = 0;
  PORTF = 0;
}

void out_addr_data(uint32_t addr, uint16_t data) {
  out(CTL_IPE, (addr >> 16) & 255);
  out(CTL_IPH, (addr >> 8) & 255);
  out(CTL_IPL, addr & 255);
  out(CTL_WH, (data >> 8) & 255);
  out(CTL_WL, data & 255);
}

//---

#define P_CE 41
#define P_OE 40
#define P_WE 39

uint8_t read_byte(uint32_t addr) {
  digitalWriteFast(P_CE, LOW);
  digitalWriteFast(P_OE, LOW);

  PORTC = addr & 255;
  PORTA = (addr >> 8) & 255;
  PORTB = (PORTB & 0b11111000) | ((addr >> 16) & 7);
  //digitalWrite(53, (addr >> 16) & 1);
  //digitalWrite(52, (addr >> 17) & 1);
  //digitalWrite(51, (addr >> 18) & 1);

  delayMicroseconds(2);

  uint8_t val = PINL;

  digitalWriteFast(P_CE, HIGH);
  digitalWriteFast(P_OE, HIGH);

  return val;
}

void write_addr_data(uint32_t addr, uint8_t data) {
  PORTC = addr & 255;
  PORTA = (addr >> 8) & 255;
  PORTB = (PORTB & 0b11111000) | ((addr >> 16) & 7);
  //digitalWrite(53, (addr >> 16) & 1);
  //digitalWrite(52, (addr >> 17) & 1);
  //digitalWrite(51, (addr >> 18) & 1);

  PORTL = data;

  delayMicroseconds(2);

  digitalWriteFast(P_WE, LOW);
  //delayMicroseconds(1);
  digitalWriteFast(P_WE, HIGH);
  //delayMicroseconds(1);
}

void write_byte(uint32_t addr, uint8_t val) {
  digitalWriteFast(P_CE, LOW);
  digitalWriteFast(P_OE, HIGH);
  DDRL = 255;

  write_addr_data(0x5555UL, 0xAA);
  write_addr_data(0x2AAAUL, 0x55);
  write_addr_data(0x5555UL, 0xA0);
  write_addr_data(addr, val);

  DDRL = 0;
  PORTL = 0;

  digitalWriteFast(P_OE, LOW);
  while ((PINL & 128) != (val & 128)) {}
  digitalWriteFast(P_OE, HIGH);

  digitalWriteFast(P_CE, HIGH);
}

void sector_erase(uint32_t addr) {
  digitalWriteFast(P_CE, LOW);
  digitalWriteFast(P_OE, HIGH);
  DDRL = 255;

  write_addr_data(0x5555UL, 0xAA);
  write_addr_data(0x2AAAUL, 0x55);
  write_addr_data(0x5555UL, 0x80);
  write_addr_data(0x5555UL, 0xAA);
  write_addr_data(0x2AAAUL, 0x55);
  write_addr_data(addr, 0x30);

  DDRL = 0;

  digitalWriteFast(P_OE, LOW);
  while ((PINL & 128) == 0) {}
  digitalWriteFast(P_OE, HIGH);

  digitalWriteFast(P_CE, HIGH);
}

void setup() {
  pinMode(CTL_IPH, OUTPUT); digitalWrite(CTL_IPH, LOW);
  pinMode(CTL_IPL, OUTPUT); digitalWrite(CTL_IPL, LOW);
  digitalWrite(CTL_SEG, HIGH); pinMode(CTL_SEG, OUTPUT);
  digitalWrite(CTL_IPE, HIGH); pinMode(CTL_IPE, OUTPUT);

  digitalWrite(CTL_WH, HIGH); pinMode(CTL_WH, OUTPUT);
  digitalWrite(CTL_WL, HIGH); pinMode(CTL_WL, OUTPUT);

  digitalWrite(CTL_WH_EN, HIGH); pinMode(CTL_WH_EN, OUTPUT);
  digitalWrite(CTL_WL_EN, HIGH); pinMode(CTL_WL_EN, OUTPUT);

  digitalWrite(CTL_RH, HIGH); pinMode(CTL_RH, OUTPUT);
  digitalWrite(CTL_RL, HIGH); pinMode(CTL_RL, OUTPUT);

  pinMode(BTN_WRITE, INPUT_PULLUP);
  pinMode(BTN_READ, INPUT_PULLUP);
  pinMode(LED_EXT, OUTPUT); digitalWrite(LED_EXT, LOW);
  digitalWrite(HALT_EXT, HIGH); pinMode(HALT_EXT, OUTPUT);

  DDRF = 0;
  PORTF = 0;

  pinMode(FLASHSEL_B1, INPUT_PULLUP);
  pinMode(FLASHSEL_B0, INPUT_PULLUP);
  
  //---

  Serial.begin(230400);
  //Serial.begin(115200);

  DDRC = 255;
  DDRA = 255;
  DDRB |= 7;
  //pinMode(51, OUTPUT);
  //pinMode(52, OUTPUT);
  //pinMode(53, OUTPUT);
  DDRL = 0;

  digitalWrite(P_CE, HIGH); pinMode(P_CE, OUTPUT);
  digitalWrite(P_OE, HIGH); pinMode(P_OE, OUTPUT);
  digitalWrite(P_WE, HIGH); pinMode(P_WE, OUTPUT);

  while (Serial.available()) {
    Serial.read();
  }
}

void loop() {
  if(digitalRead(FLASHSEL_B1) == HIGH) {
    //0 bit
    flashsel_offset &= ~(262144UL);
  } else {
    //1 bit
    flashsel_offset |= 262144UL;
  }
  
  if(digitalRead(FLASHSEL_B0) == HIGH) {
    //0 bit
    flashsel_offset &= ~(131072UL);
  } else {
    //1 bit
    flashsel_offset |= 131072UL;
  }
  
  if (Serial.available()) {
    char ch = Serial.read();
    if (ch == 'w' || ch == 'r') {
      delayMicroseconds(3000 / 2); //more than enough for 9 bytes
      uint32_t start = ((uint32_t)Serial.read() << 24) |
                       ((uint32_t)Serial.read() << 16) |
                       ((uint32_t)Serial.read() << 8) |
                       ((uint32_t)Serial.read());

      uint32_t len = ((uint32_t)Serial.read() << 24) |
                     ((uint32_t)Serial.read() << 16) |
                     ((uint32_t)Serial.read() << 8) |
                     ((uint32_t)Serial.read());

      if (ch == 'w') {
        for (uint32_t i = 0; i < len; i++) {
          if (!Serial.available()) {
            delayMicroseconds(600 / 2);
          }
          if (!Serial.available()) {
            break;
          }

          uint8_t val = Serial.read();

          uint32_t idx = start + i;
          //dunno why FIXME
          //if(idx == 65536) { idx = 0; } else if(idx == 0) { idx = 65536; }
          write_byte(flashsel_offset + idx, val);
        }
      } else if (ch == 'r') {
        for (uint32_t i = 0; i < len; i++) {
          Serial.write(read_byte(flashsel_offset + start + i));
        }
      }
    } else if (ch == 'l') {
      delayMicroseconds(1500 / 2);

      //write_byte(flashsel_offset + 0x1fff0, Serial.read());
      //write_byte(flashsel_offset + 0x1fff1, Serial.read());
      //write_byte(flashsel_offset + 0x1fff2, Serial.read());
      //write_byte(flashsel_offset + 0x1fff3, Serial.read());
    } else if (ch == 'e') {
      delayMicroseconds(3000 / 2); //more than enough for 9 bytes
      uint32_t start = ((uint32_t)Serial.read() << 24) |
                       ((uint32_t)Serial.read() << 16) |
                       ((uint32_t)Serial.read() << 8) |
                       ((uint32_t)Serial.read());

      uint32_t len = ((uint32_t)Serial.read() << 24) |
                     ((uint32_t)Serial.read() << 16) |
                     ((uint32_t)Serial.read() << 8) |
                     ((uint32_t)Serial.read());
      
      for (uint32_t i = start; i < start + len + 4096; i += 4096) {
        sector_erase(flashsel_offset + (i & 0b11111111111111111111000000000000UL));
      }
      Serial.write('d');
    }

    //ignore the rest
    delayMicroseconds(300 / 2);
    while (Serial.available()) {
      while (Serial.available()) {
        Serial.read();
      }
      delayMicroseconds(300 / 2);
    }
  }

  if (digitalRead(BTN_WRITE) == LOW) {
    while (digitalRead(BTN_WRITE) == LOW) {}

    digitalWrite(HALT_EXT, LOW);
    digitalWrite(LED_EXT, HIGH);

    uint32_t data_len = ((uint32_t)read_byte(flashsel_offset + 0x1fff0) << 24) |
                        ((uint32_t)read_byte(flashsel_offset + 0x1fff1) << 16) |
                        ((uint32_t)read_byte(flashsel_offset + 0x1fff2) << 8) |
                        ((uint32_t)read_byte(flashsel_offset + 0x1fff3));
    //Serial.println(data_len);
    //data_len = 262144UL;
    //data_len = 98304UL;
    //data_len = 262144UL;
    //data_len = 327680UL;
    //data_len = 393216UL;
    //data_len = 524288UL;
    //data_len = 8192UL;
    data_len = 32768UL;
    data_len >>= 1;

#if defined W_RAM
    out(CTL_SEG, 0b10001000);
    
    for(uint32_t index = 0; index < data_len; index++) {
      if (Serial.available()) {
        break;
      }
      
      out(CTL_IPE, (index >> 16) & 255);
      out(CTL_IPH, (index >> 8) & 255);
      out(CTL_IPL, index & 255);
      out(CTL_WH, read_byte( flashsel_offset + ((index << 1) | 0UL) ));
      out(CTL_WL, read_byte( flashsel_offset + ((index << 1) | 1UL) ));
    }
#elif defined W_FLASH
    out(CTL_SEG, 0);

    //erase flash
    for (uint32_t index = 0; index < data_len; index += 4096) {
      out_addr_data(0x5555UL, 0xAAAAU);
      out_addr_data(0x2AAAUL, 0x5555U);
      out_addr_data(0x5555UL, 0x8080U);
      out_addr_data(0x5555UL, 0xAAAAU);
      out_addr_data(0x2AAAUL, 0x5555U);
      out_addr_data(index, 0x3030U);
      delay(25);
    }

    //write flash
    for(uint32_t index = 0; index < data_len; index++) {
      if (Serial.available()) {
        break;
      }

      uint16_t data = 0;
      data |= ((uint16_t)read_byte( flashsel_offset + ((index << 1) | 0UL) )) << 8;
      data |= (uint16_t)read_byte( flashsel_offset + ((index << 1) | 1UL) );

      out_addr_data(0x5555UL, 0xAAAAU);
      out_addr_data(0x2AAAUL, 0x5555U);
      out_addr_data(0x5555UL, 0xA0A0U);
      out_addr_data(index, data);
      delayMicroseconds(25);
    }
#endif

    out(CTL_IPE, 0);
    out(CTL_IPH, 0);
    out(CTL_IPL, 0);

    digitalWrite(LED_EXT, LOW);
    digitalWrite(HALT_EXT, HIGH);
  }

  if (digitalRead(BTN_READ) == LOW) {
    while (digitalRead(BTN_READ) == LOW) {}

    digitalWrite(HALT_EXT, LOW);
    digitalWrite(LED_EXT, HIGH);

    uint32_t data_len = ((uint32_t)read_byte(flashsel_offset + 0x1fff0) << 24) |
                        ((uint32_t)read_byte(flashsel_offset + 0x1fff1) << 16) |
                        ((uint32_t)read_byte(flashsel_offset + 0x1fff2) << 8) |
                        ((uint32_t)read_byte(flashsel_offset + 0x1fff3));
    //Serial.println(data_len);
    //data_len = 262144UL;
    //data_len = 393216UL;
    //data_len = 524288UL;
    data_len = 32768UL;
    data_len >>= 1;

    for (uint32_t i = 0; i < (data_len << 1); i += 4096) {
      sector_erase(flashsel_offset + i);
    }

#if defined W_RAM
    out(CTL_SEG, 0b10001000);
#elif defined W_FLASH
    out(CTL_SEG, 0);
#endif

    for(uint32_t index = 0; index < data_len; index++) {
      if (Serial.available()) {
        break;
      }
      
      out(CTL_IPE, (index >> 16) & 255);
      out(CTL_IPH, (index >> 8) & 255);
      out(CTL_IPL, index & 255);

      digitalWriteFast(CTL_RH, LOW);
      delayMicroseconds(5);
      write_byte( flashsel_offset + ((index << 1) | 0UL) , PINF);
      digitalWriteFast(CTL_RH, HIGH);
      digitalWriteFast(CTL_RL, LOW);
      delayMicroseconds(5);
      write_byte( flashsel_offset + ((index << 1) | 1UL) , PINF);
      digitalWriteFast(CTL_RL, HIGH);
    }

    out(CTL_IPE, 0);
    out(CTL_IPH, 0);
    out(CTL_IPL, 0);

    digitalWrite(LED_EXT, LOW);
    digitalWrite(HALT_EXT, HIGH);
  }
}
